/*
 * linux/arch/m68k/atari/mach64fb.c -- Low level implementation of the
 *                                    ATI Mach64 frame buffer device
 *
 *    Copyright (C) 1997 Bernd Harries
 *
 *
 * This file is based on the Amiga frame buffer device (amifb.c):
 *
 *    Copyright (C) 1995 Geert Uytterhoeven
 *
 *
 * History:
 *   - 22 Dec 95: Original version by Martin Apel
 *   - 05 Jan 96: Geert: integration into the current source tree
 *
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 */

/*============================================================================*/
/* Includes */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/zorro.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/pgtable.h>

#include "mach64regs.h"

#include "mach64fb.h"


/*============================================================================*/
/* Defines */

#define _MAP_768K_IO_

#define _NOT_DUMP_IO_REGS_

#define _NOT_DUMP_GUIREGS_0BFC00_

#define _NOT_DUMP_GUIREGS_3FFC00_

#define _NOT_MAP_GUIREGS_0BFC00_


#ifdef MULTI_MACH

#define MACH_MAX 8

#define DECLARE_STATIC_UINT08___MN_  static unsigned char          m64_num;
#define DECLARE_UINT08___MN_         unsigned char m64_num;
#define ARG_UINT08___MN_             unsigned char m64_num
#define ADD_ARG_UINT08___MN_         , unsigned char m64_num

#define _ARRAY_       [MACH_MAX]
#define __MN__        [m64_num]
#define _MN_           m64_num
#define _INC_MN_       m64_num++
#define _ARG_MN_       m64_num
#define _ADD_ARG_MN_   , m64_num

#else      /*---------------------*/

#define DECLARE_STATIC_UINT08___MN_
#define DECLARE_UINT08___MN_
#define ARG_UINT08___MN_           void
#define ADD_ARG_UINT08___MN_

#define _ARRAY_    
#define __MN__         
#define _MN_           0
#define _INC_MN_       
#define _ARG_MN_       
#define _ADD_ARG_MN_   

#endif     /* MULTI_MACH */

/*--------------------------------------------------------------------------*/
  
#define MACH_MAX 8
#define arraysize(x)    (sizeof(x)/sizeof(*(x)))

#define  FuncId       if(mach64fb_debug & 0x00000100) printk
#define  FuncDone     if(mach64fb_debug & 0x00000200) printk
#define  CFuncId      if(mach64fb_debug & 0x00000400) printk
#define  CFuncDone    if(mach64fb_debug & 0x00000800) printk
#define  OFuncId      if(mach64fb_debug & 0x00001000) printk
#define  OFuncDone    if(mach64fb_debug & 0x00002000) printk


/* Busy */
#define  Bprint       if(mach64fb_debug & 0x01000000) printk
/* Debug */
#define  Dprint       if(mach64fb_debug & 0x00800000) printk

/* Message */
#define  Mprint       if(mach64fb_debug & 0x00004000) printk
/* Info */
#define  Iprint       if(mach64fb_debug & 0x00008000) printk
/* Warn */
#define  Wprint       if(mach64fb_debug & 0x00010000) printk
/* Error */
#define  Eprint       if(mach64fb_debug & 0x00020000) printk
/* Fatal */
#define  Fprint       if(mach64fb_debug & 0x00040000) printk
/* Panic */
#define  Pprint       if(mach64fb_debug & 0x00080000) printk


/*--------------------------------------------------------------------------*/

#ifdef __alpha__
/*
 * on the Alpha, there are "write buffers" that hold data to be written
 *  to memory; the data is *not* necessarily flushed to memory immediately.
 * so, we use a "memory barrier" instruction to force the flush of these
 *  buffers, so that writing to the memory-mapped Mach64 registers *will*
 *  take place immediately.
 */

#define barrier() __asm__ __volatile__("mb": : :"memory")

#else /* __alpha__ */

/*
#define barrier()
*/

#endif /* __alpha__ */

/*--------------------------------------------------------------------------*/

#ifdef _MAP_768K_IO_

#define OUTB(port,val) \
  *((unsigned volatile char *) ((port) + 0x00080000 + virt_vgaiobase __MN__  )) = (val)

#define INB(port) \
  (*((unsigned volatile char *) ((port) + 0x00080000 + virt_vgaiobase __MN__  )))

/*
#define OUTL(port,val) \
  *((unsigned volatile int *) ((port) + 0x00080000 + virt_vgaiobase __MN__  )) = (val)

#define INL(port) \
  (*((unsigned volatile int *) ((port) + 0x00080000 + virt_vgaiobase __MN__  )))
*/

#else

#define OUTB(port,val) \
  *((unsigned volatile char *) ((port) + virt_vgaiobase __MN__  )) = (val)

#define INB(port) \
  (*((unsigned volatile char *) ((port) + virt_vgaiobase __MN__  )))

/*
#define OUTL(port,val) \
  *((unsigned volatile int *) ((port) + virt_vgaiobase __MN__  )) = (val)

#define INL(port) \
  (*((unsigned volatile int *) ((port) + virt_vgaiobase __MN__  )))
*/

#endif  /* _MAP_768K_IO_ */

#define DACDelay                    \
  do {                              \
    unsigned char tmp = INB(0x3da); \
    tmp = INB(0x3da);               \
  } while (0)

/*============================================================================*/

/*
#define regwr(port, val) \
  *((unsigned volatile int *) ((port) + virt_vgaiobase __MN__  )) = (val)


#define regrd(port) \
  (*((unsigned volatile int *) ((port) + virt_vgaiobase __MN__  )))
*/

/*============================================================================*/
/* Types */

struct mach64_fb_par
{
   int xres;
   int yres;
   int bpp;
};

/*============================================================================*/
/* Extern Functions*/



/*============================================================================*/
/* Intern Functions*/

/*--------------------------------------------------------------------------*/
   /*
    *    Interface to the low level console driver exported with fb_info
    */

static int  Mach64fb_switch(int con);
static int  Mach64fb_updatevar(int con);
static void Mach64fb_blank(int blank);
static int  Mach64fb_setcmap(struct fb_cmap * cmap, int con);

/*--------------------------------------------------------------------------*/
   /*
    *    Interface fb_ops used by the world exported with fb_info.fb_ops
    */

static int M64_op_open(int fbidx);
static int M64_op_release(int fbidx);
static int M64_op_get_fix(struct fb_fix_screeninfo * fix, int con);
static int M64_op_get_var(struct fb_var_screeninfo * var, int con);
static int M64_op_set_var(struct fb_var_screeninfo * var, int con);
static int M64_op_get_cmap(struct fb_cmap * cmap, int kspc, int con);
static int M64_op_set_cmap(struct fb_cmap * cmap, int kspc, int con);
static int M64_op_pan_display(struct fb_var_screeninfo * var, int con);

static int 
M64_op_ioctl(
  struct inode * inode,
  struct file *file,
  u_int cmd,
  u_long arg,
  int con);

/*--------------------------------------------------------------------------*/
   /*
    *   Internal Routines
    */

static int 
do_fb_set_var(struct fb_var_screeninfo * var, int isactive);

static void 
do_install_cmap(int con);

static void 
Mach64_fb_set_disp(int con);

static int 
get_video_mode(const char * name);

static int 
store_video_par(char* videopar  ADD_ARG_UINT08___MN_ );

static char * 
strtoke(char * s,const char * ct);

/*--------------------------------------------------------------------------*/
   /*
    *   Hardware Specific Routines
    */

static int 
M64_hw_init( ARG_UINT08___MN_ );

static int 
M64_hw_encode_fix(struct fb_fix_screeninfo * fix, struct mach64_fb_par * par);

static int 
M64_hw_decode_var(struct fb_var_screeninfo * var, struct mach64_fb_par * par);

static int 
M64_hw_encode_var(struct fb_var_screeninfo * var, struct mach64_fb_par * par);

static void 
M64_hw_get_par(struct mach64_fb_par * par);

static void 
M64_hw_set_par(struct mach64_fb_par * par);

static int 
M64_hw_getcolreg(
  u_int regno,
  u_int * red,
  u_int * green,
  u_int * blue,
  u_int * transp);

static int 
M64_hw_setcolreg(u_int regno, u_int red, u_int green, u_int blue, u_int transp);

static void 
M64_hw_blank(int blank);



/*============================================================================*/
/*============================================================================*/
/* Intern Variables */

/*--------------------------------------------------------------------------*/
/* BSS */


static unsigned long            phys_vmembase _ARRAY_ ;
static unsigned long            phys_size _ARRAY_ ;
static unsigned long            phys_vgaiobase _ARRAY_ ;
static unsigned long            phys_mmregbase _ARRAY_ ;

static unsigned long            virt_vmembase _ARRAY_ ;
static unsigned long            virt_size _ARRAY_ ;
static unsigned long            virt_vgaiobase _ARRAY_ ;
static unsigned long            virt_mmregbase _ARRAY_ ;

static unsigned int             mach64_xres _ARRAY_ ;
static unsigned int             mach64_yres _ARRAY_ ;
static unsigned int             mach64_depth _ARRAY_ ;


static struct mach64_fb_par     current_par _ARRAY_ ;
static struct display           m_disp _ARRAY_ ;
static struct fb_info           fb_info _ARRAY_ ;

static u_short                  mach64_colormap  _ARRAY_ [256][4];

static long *                   memstart;

static u_int                    mach64_key;
static u_int                    mach64_bitspercol _ARRAY_ ;

static int                      currcon;


static u_short                  mach64_inverse _ARRAY_ ;
static short int                mach64_pmode _ARRAY_ ;
static short int                mach64_mode _ARRAY_ ;   /* 0..31 */
static short int                mach64_use_accel _ARRAY_ ;
static char                     current_par_valid _ARRAY_ ;

static unsigned char            mach64_count;


/*--------------------------------------------------------------------------*/
   /*
    *    Switch for Chipset Independency
    */

static struct fb_hwswitch
{

  /* Initialisation */

  int (*init)( ARG_UINT08___MN_ );

  /* Display Control */

  int (*encode_fix)(struct fb_fix_screeninfo * fix, struct mach64_fb_par * par);
  int (*decode_var)(struct fb_var_screeninfo * var, struct mach64_fb_par * par);
  int (*encode_var)(struct fb_var_screeninfo * var, struct mach64_fb_par * par);
  void (*get_par)(struct mach64_fb_par * par );
  void (*set_par)(struct mach64_fb_par * par );
  int (*getcolreg)(u_int regno, u_int * red, u_int * green, u_int * blue,
    u_int * transp);
  int (*setcolreg)(u_int regno, u_int red, u_int green, u_int blue,
    u_int transp);
  void (*blank)(int blank);
  void (*set_screen_base)( unsigned long s_base );
  int  (*pan_display)( struct fb_var_screeninfo * var);
} 
* fbhws_ptr;


/*--------------------------------------------------------------------------*/
/* Data */


/* -------------- local Interfaces to hardware functions ---------------- */

/* see XGCValues GCFunction */
int 
mach64alu[16] =
{
    MIX_0,
    MIX_AND,
    MIX_SRC_AND_NOT_DST,
    MIX_SRC,
    MIX_NOT_SRC_AND_DST,
    MIX_DST,
    MIX_XOR,
    MIX_OR,
    MIX_NOR,
    MIX_XNOR,
    MIX_NOT_DST,
    MIX_SRC_OR_NOT_DST,
    MIX_NOT_SRC,
    MIX_NOT_SRC_OR_DST,
    MIX_NAND,
    MIX_1
};

static struct fb_hwswitch 
mach64_switch = 
{
  M64_hw_init,
  M64_hw_encode_fix,
  M64_hw_decode_var,
  M64_hw_encode_var,
  NULL,                    /* M64_hw_get_par, */
  NULL,                    /* M64_hw_set_par, */
  M64_hw_getcolreg,
  M64_hw_setcolreg,
  M64_hw_blank,
  NULL,
  NULL
};

/* ------ Interfaces to hardware functions to export with fb_info  ------ */

static struct fb_ops 
mach64_fb_ops =
{
  M64_op_open,
  M64_op_release,
  M64_op_get_fix,
  M64_op_get_var,
  M64_op_set_var,
  M64_op_get_cmap,
  M64_op_set_cmap,
  M64_op_pan_display,
  M64_op_ioctl
};

static  unsigned int    mach64fb_debug = 0x018FCF00;

   /*
    *    Frame Buffer Name
    */

static char     mach64_fb_name[16] = "ATI Mach64 ";


   /*
    *    Ati Mach64 Graphics Board
    */

#define M64_8_WIDTH 1152
#define M64_8_HEIGHT 912
#define M64_8_PIXCLOCK 12500    /* ++Geert: Just a guess */

#define M64_16_WIDTH 800
#define M64_16_HEIGHT 600
#define M64_16_PIXCLOCK 25000   /* ++Geert: Just a guess */


/*--------------------------------------------------------------------------*/
   /*
    *    Predefined Video Mode Names
    */
/*--------------------------------------------------------------------------*/

static char *                   fb_var_names[] =
{

   /*
    *    Autodetect (Default) Video Mode
    */

   "autodetect",

   /*
    *    Predefined Video Modes
    */
    
   "Mach64_8",            /* Mach64 8 bpp */
   "Mach64_16",           /* Mach64 16 bpp */

   /*
    *    Dummy Video Modes
    */

   "dummy", "dummy", "dummy", "dummy", "dummy", "dummy", "dummy", "dummy",
   "dummy", "dummy", "dummy", "dummy", "dummy", "dummy", "dummy", "dummy",
   "dummy", "dummy", "dummy", "dummy",

   /*
    *    User Defined Video Modes
    *
    *    This doesn't work yet!!
    */

   "user0", "user1", "user2", "user3", "user4", "user5", "user6", "user7"
};


/*--------------------------------------------------------------------------*/
   /*
    *    Predefined Video Mode Definitions
    */
/*--------------------------------------------------------------------------*/

static struct fb_var_screeninfo mach64_fb_predefined[] = 
{

   /*
    *    Autodetect (Default) Video Mode
    */

   { 0, },

   /*
    *    Predefined Video Modes
    */
    
   {
      /* Cybervision 8 bpp */
      M64_8_WIDTH, M64_8_HEIGHT, M64_8_WIDTH, M64_8_HEIGHT, 0, 0, 8, 0,
      {0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
      0, 0, -1, -1, FB_ACCEL_NONE, M64_8_PIXCLOCK, 64, 96, 35, 12, 112, 2,
      FB_SYNC_COMP_HIGH_ACT|FB_SYNC_VERT_HIGH_ACT, FB_VMODE_NONINTERLACED
   }, {
      /* Cybervision 16 bpp */
      M64_16_WIDTH, M64_16_HEIGHT, M64_16_WIDTH, M64_16_HEIGHT, 0, 0, 16, 0,
      {11, 5, 0}, {5, 6, 0}, {0, 5, 0}, {0, 0, 0},
      0, 0, -1, -1, FB_ACCEL_NONE, M64_16_PIXCLOCK, 64, 96, 35, 12, 112, 2,
      FB_SYNC_COMP_HIGH_ACT|FB_SYNC_VERT_HIGH_ACT, FB_VMODE_NONINTERLACED
   },

   /*
    *    Dummy Video Modes
    */

   { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, },
   { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, },
   { 0, }, { 0, },

   /*
    *    User Defined Video Modes
    */

   { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, }, { 0, }
};


#define NUM_TOTAL_MODES    arraysize(mach64_fb_predefined)
#define NUM_PREDEF_MODES   (3)


   /*
    *    Some default modes
    */

#define M64_8_DEFMODE     (1)
#define M64_16_DEFMODE    (2)



/*============================================================================*/
/*============================================================================*/
/* Functions */

/*============================================================================*/
/*============================================================================*/


/*============================================================================*/

__inline__ void 
regwr(unsigned int regindex, unsigned long regdata)
{
  register unsigned long        appaddr;
  register u_int                lword;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((regindex) + 0x003FFC00 + virt_vmembase __MN__  );

  lword = (regdata >> 16) | (regdata << 16);
  lword = ((lword >> 8) & 0x00FF00FF) | ((lword << 8) & 0xFF00FF00);

  *((unsigned volatile int*)(appaddr)) = lword;
  barrier();
}
/*============================================================================*/

__inline__ unsigned int
regrd(unsigned int regindex)
{
  register unsigned long        appaddr;
  register u_int                lword;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((regindex) + 0x003FFC00 + virt_vmembase __MN__  );

  lword = (*((unsigned volatile int*)(appaddr)));

  lword = (lword >> 16) | (lword << 16);     /* swap words */
  lword = ((lword >> 8) & 0x00FF00FF) | ((lword << 8) & 0xFF00FF00);

  return(lword);
}
/*============================================================================*/

__inline__ void 
regwrb(unsigned int regindex, unsigned char regdata)
{
  register unsigned long        appaddr;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((regindex) + 0x003FFC00 + virt_vmembase __MN__  );

  *(char *)appaddr = regdata;
  barrier();
}
/*============================================================================*/


#ifdef _MAP_768K_IO_

__inline__ void
OUTL(unsigned int port, unsigned long regdata)
{
  register unsigned long        appaddr;
  register u_int                lword;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((port) + 0x00080000 + virt_vgaiobase __MN__  );

  lword = (regdata >> 16) | (regdata << 16);
  lword = ((lword >> 8) & 0x00FF00FF) | ((lword << 8) & 0xFF00FF00);

  *((unsigned volatile int*)(appaddr)) = lword;
  barrier();
}
/*============================================================================*/

__inline__ unsigned int
INL(unsigned int port)
{
  register unsigned long        appaddr;
  register u_int                lword;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((port) + 0x00080000 + virt_vgaiobase __MN__  );

  lword = (*((unsigned volatile int*)(appaddr)));

  lword = (lword >> 16) | (lword << 16);     /* swap words */
  lword = ((lword >> 8) & 0x00FF00FF) | ((lword << 8) & 0xFF00FF00);

  return(lword);
}
/*============================================================================*/

__inline__ void 
regw(unsigned int regindex, unsigned long regdata)
{
  register unsigned long        appaddr;
  register u_int                lword;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((regindex) + 0x0003FC00 + virt_vgaiobase __MN__  );

  lword = (regdata >> 16) | (regdata << 16);
  lword = ((lword >> 8) & 0x00FF00FF) | ((lword << 8) & 0xFF00FF00);

  *((unsigned volatile int*)(appaddr)) = lword;
  barrier();
}
/*============================================================================*/

__inline__ unsigned int
regr(unsigned int regindex)
{
  register unsigned long        appaddr;
  register u_int                lword;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((regindex) + 0x0003FC00 + virt_vgaiobase __MN__  );

  lword = (*((unsigned volatile int*)(appaddr)));

  lword = (lword >> 16) | (lword << 16);     /* swap words */
  lword = ((lword >> 8) & 0x00FF00FF) | ((lword << 8) & 0xFF00FF00);

  return(lword);
}
/*============================================================================*/

__inline__ void 
regwB(unsigned int regindex, unsigned char regdata)
{
  register unsigned long        appaddr;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((regindex) + 0x0003FC00 + virt_vgaiobase __MN__  );

  *(char *)appaddr = regdata;
  barrier();
}
/*============================================================================*/

#else

__inline__ void 
OUTL(unsigned int port, unsigned long regdata)
{
  register unsigned long        appaddr;
  register u_int                lword;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((port) + virt_vgaiobase __MN__  );

  lword = (regdata >> 16) | (regdata << 16);
  lword = ((lword >> 8) & 0x00FF00FF) | ((lword << 8) & 0xFF00FF00);

  *((unsigned volatile int*)(appaddr)) = lword;
  barrier();
}
/*============================================================================*/

__inline__ unsigned int
INL(unsigned int port)
{
  register unsigned long        appaddr;
  register u_int                lword;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((port) + virt_vgaiobase __MN__  );

  lword = (*((unsigned volatile int*)(appaddr)));

  lword = (lword >> 16) | (lword << 16);     /* swap words */
  lword = ((lword >> 8) & 0x00FF00FF) | ((lword << 8) & 0xFF00FF00);

  return(lword);
}
/*============================================================================*/



__inline__ void 
regw(unsigned int regindex, unsigned long regdata)
{
  register unsigned long        appaddr;
  register u_int                lword;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((regindex) + 0x0000FC00 + virt_mmregbase __MN__  );

  lword = (regdata >> 16) | (regdata << 16);
  lword = ((lword >> 8) & 0x00FF00FF) | ((lword << 8) & 0xFF00FF00);

  *((unsigned volatile int*)(appaddr)) = lword;
  barrier();
}
/*============================================================================*/

__inline__ unsigned int
regr(unsigned int regindex)
{
  register unsigned long        appaddr;
  register u_int                lword;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((regindex) + 0x0000FC00 + virt_mmregbase __MN__  );

  lword = (*((unsigned volatile int*)(appaddr)));

  lword = (lword >> 16) | (lword << 16);     /* swap words */
  lword = ((lword >> 8) & 0x00FF00FF) | ((lword << 8) & 0xFF00FF00);

  return(lword);
}
/*============================================================================*/

__inline__ void 
regwB(unsigned int regindex, unsigned char regdata)
{
  register unsigned long        appaddr;
  /*--------------------------------------------------------------------------*/
  /* calculate aperture address */
  appaddr = ((regindex) + 0x0000FC00 + virt_mmregbase __MN__  );

  *(char *)appaddr = regdata;
  barrier();
}
/*============================================================================*/

#endif  /* _MAP_768K_IO_ */


/*============================================================================*/
/*============================================================================*/


/*--------------------------------------------------------------------------*/

static int 
M64_op_open(int fbidx)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  OFuncId("M64_op_open(%02d) \n", fbidx);
  /*
   * Nothing, only a usage count for the moment
   */

  MOD_INC_USE_COUNT;

  return(0);
}
/*endproc M64_op_open() */

/*============================================================================*/

/*--------------------------------------------------------------------------*/

static int 
M64_op_release(int fbidx)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  FuncId("M64_op_release() \n");
  MOD_DEC_USE_COUNT;
  return(0);
}
/*endproc M64_op_release() */

/*============================================================================*/


/* -------------------- Hardware specific routines ------------------------ */


/*--------------------------------------------------------------------------*/
   /*
    *    Initialization
    *
    *    Set the default video mode for this chipset. If a video mode was
    *    specified on the command line, it will override the default mode.
    */
/*--------------------------------------------------------------------------*/

static int 
M64_hw_init( ARG_UINT08___MN_ )

/*--------------------------------------------------------------------------*/
{
  struct fb_var_screeninfo *    var_ptr = &mach64_fb_predefined[0];
  struct mach64_fb_par          dummy_par;

  unsigned long                 board_addr;
  u_int *                       u32_ptr;
  u_int                         read_val;
  int                           n;
  int                           i;

  /*--------------------------------------------------------------------------*/
  FuncId("M64_hw_init() \n");

  if(mach64_mode __MN__   == -1)
  {
    mach64_mode __MN__   = M64_8_DEFMODE;
  }
  /*endif*/

  board_addr = virt_vmembase __MN__  ;
  for (i = 0; i < 256; i++)
  {
    mach64_colormap  __MN__  [i][0] = i;
    mach64_colormap  __MN__  [i][1] = i;
    mach64_colormap  __MN__  [i][2] = i;
    mach64_colormap  __MN__  [i][3] = 0;
  }

#ifdef _DUMP_IO_REGS_
  u32_ptr = (u_int *)(0x000002EC + virt_vgaiobase __MN__ );
  n = 0;
  while(n < 0x20)
  {
    Dprint("\n%08X: ", u32_ptr);
    Dprint("%08X ", INL(0x02EC + (n << 10)) );  n++;
    Dprint("%08X ", INL(0x02EC + (n << 10)) );  n++;
    Dprint("%08X ", INL(0x02EC + (n << 10)) );  n++;
    Dprint("%08X ", INL(0x02EC + (n << 10)) );  n++;
    u32_ptr += 0x0400;
  }
  /*endwhile*/
  Dprint("\n" );
#endif

  WaitIdleEmpty(); /* Make sure that all commands have finished */
  read_val = regr(GEN_TEST_CNTL);
  if(0)
  {
    regw(GEN_TEST_CNTL, read_val | 0x0200);
  }
  else
  {
    regw(GEN_TEST_CNTL, read_val & ~0x0200);
  }
  /*endif*/

  WaitQueue(13);
  regw(SRC_OFF_PITCH, (mach64_xres __MN__  << 19));
  regw(DST_OFF_PITCH, (mach64_xres __MN__  << 19));
  regw(SC_LEFT_RIGHT, (mach64_xres __MN__  -1) << 16);
  regw(SC_TOP_BOTTOM, (mach64_yres __MN__  -1) << 16);

  regw(DP_BKGD_CLR, 0x00000000);
  regw(DP_FRGD_CLR, 0x00000001);

  regw(DP_WRITE_MASK, 0x000000FF  /* pGC->planemask */ );
  regw(DP_PIX_WIDTH, BYTE_ORDER_MSB_TO_LSB | HOST_8BPP | SRC_8BPP | DST_8BPP);
  regw(DP_MIX, FRGD_MIX_S | BKGD_MIX_D);
  regw(DP_SRC, FRGD_SRC_BLIT | BKGD_SRC_BKGD_CLR);

  regw(CLR_CMP_CNTL, 0x00000000);
  regw(CLR_CMP_CLR, 0x00000000);

  regw(GUI_TRAJ_CNTL, DST_X_LEFT_TO_RIGHT | DST_Y_TOP_TO_BOTTOM);



#ifdef  _DUMP_GUIREGS_0BFC00_

#ifdef _MAP_768K_IO_
  u32_ptr = (u_int *)(0x0003FC00 + virt_vgaiobase __MN__ );
#else
  u32_ptr = (u_int *)(0x0000FC00 + virt_mmregbase __MN__ );
#endif
  n = 0;
  while(n < 0x100)
  {
    Dprint("\n%08X: ", u32_ptr);
    Dprint("%08X ", regr(n << 2) );  n++;
    Dprint("%08X ", regr(n << 2) );  n++;
    Dprint("%08X ", regr(n << 2) );  n++;
    Dprint("%08X ", regr(n << 2) );  n++;
    u32_ptr += 4;
  }
  /*endwhile*/
  Dprint("\n" );
#endif

#ifdef  _DUMP_GUIREGS_3FFC00_
  u32_ptr = (u_int *)(0x003FFC00 + virt_vmembase __MN__ );
  n = 0;
  while(n < 0x100)
  {
    Dprint("\n%08X: ", u32_ptr);
    Dprint("%08lX ", regrd(n << 2) );  n++;
    Dprint("%08lX ", regrd(n << 2) );  n++;
    Dprint("%08lX ", regrd(n << 2) );  n++;
    Dprint("%08lX ", regrd(n << 2) );  n++;
    u32_ptr += 4;
  }
  /*endwhile*/

  Dprint("\n" );
#endif



  dummy_par.xres = mach64_xres __MN__  ;
  dummy_par.yres = mach64_yres __MN__  ;
  dummy_par.bpp = mach64_depth __MN__  ;
  M64_hw_encode_var(var_ptr, &dummy_par);

  mach64fb_debug = 0x018FC100;

  M64_hw_setcolreg (255, 56, 100, 160, 0);
  M64_hw_setcolreg (254, 0, 0, 0, 0);

  FuncDone("M64_hw_init() done.\n");
  return(1);    /* detected_mode */
}
/*endproc M64_hw_init() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    This function should fill in the `fix' structure based on the
    *    values in the `par' structure.
    */
/*--------------------------------------------------------------------------*/

static int 
M64_hw_encode_fix(struct fb_fix_screeninfo * fix, struct mach64_fb_par * par)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_
  int                           i;

  /*--------------------------------------------------------------------------*/
  CFuncId("M64_hw_encode_fix() \n");

  strcpy(fix->id, mach64_fb_name);
  fix->id[15] =  _MN_ ;
  fix->smem_start = virt_vmembase __MN__  ;

  /* aufrunden */
  fix->smem_len = (phys_size __MN__   + PAGE_SIZE - 0x00200001) & PAGE_MASK;

  if(mach64_depth __MN__   == 1)
  {
    fix->type = FB_TYPE_PACKED_PIXELS;
    /* The letters 'n' and 'i' in the "video=Mach64:" stand
     * for "normal" and "inverted", rsp., in the monochrome case */
    fix->visual =
      (mach64_pmode __MN__   == FB_TYPE_INTERLEAVED_PLANES ||
      mach64_pmode __MN__   == FB_TYPE_PACKED_PIXELS) ?
      FB_VISUAL_MONO10 : FB_VISUAL_MONO01;
  }
  else   /* depth > 1 */
  {
    switch(mach64_pmode __MN__  )
    {
      /* All visuals are STATIC, because we don't know how to change
       * colors :-(
       */
      case -1:              /* truecolor */
        fix->type = FB_TYPE_PACKED_PIXELS;
        fix->visual = FB_VISUAL_TRUECOLOR;
        break;
      case FB_TYPE_PACKED_PIXELS:
        fix->type = FB_TYPE_PACKED_PIXELS;
        fix->visual = FB_VISUAL_STATIC_PSEUDOCOLOR;
        break;
      case FB_TYPE_PLANES:
        fix->type = FB_TYPE_PLANES;
        fix->visual = FB_VISUAL_STATIC_PSEUDOCOLOR;
        break;
      case FB_TYPE_INTERLEAVED_PLANES:
        fix->type = FB_TYPE_INTERLEAVED_PLANES;
        fix->type_aux = 2;
        fix->visual = FB_VISUAL_STATIC_PSEUDOCOLOR;
        break;
    }
    /*endswitch*/
  }
  /*endif*/
  fix->xpanstep = 0;
  fix->ypanstep = 0;
  fix->ywrapstep = 0;
  fix->line_length = 0;

  for(i = 0; i < arraysize(fix->reserved); i++)
    fix->reserved[i] = 0;

  CFuncId("M64_hw_encode_fix() done. \n");
  return(0);
}
/*endproc _encode_fix() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Get the video params out of `var'. If a value doesn't fit, round
    *    it up, if it's too big, return -EINVAL.
    */
/*--------------------------------------------------------------------------*/

static int 
M64_hw_decode_var(struct fb_var_screeninfo * var, struct mach64_fb_par * par)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  CFuncId("M64_hw_decode_var() \n");

  par->xres = var->xres;
  par->yres = var->yres;
  par->bpp = var->bits_per_pixel;

  return(0);
}
/*endproc _decode_var() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Fill the `var' structure based on the values in `par' and maybe
    *    other values read out of the hardware.
    */
/*--------------------------------------------------------------------------*/

static int 
M64_hw_encode_var(struct fb_var_screeninfo * var, struct mach64_fb_par * par)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 
  int                           i;

  /*--------------------------------------------------------------------------*/
  CFuncId("M64_hw_encode_var() \n");

  var->xres = par->xres;
  var->yres = par->yres;
  var->xres_virtual = par->xres;
  var->yres_virtual = par->yres;
  var->xoffset = 0;
  var->yoffset = 0;

  var->bits_per_pixel = par->bpp;
  var->grayscale = 0;

  if(par->bpp == 8)
  {
    var->red.offset = 0;
    var->red.length =
      (virt_vgaiobase __MN__   ? mach64_bitspercol __MN__  : 0);
    var->red.msb_right = 0;
    var->blue = var->green = var->red;
  }
  else
  {
    var->red.offset = 11;
    var->red.length = 5;
    var->red.msb_right = 0;
    var->green.offset = 5;
    var->green.length = 6;
    var->green.msb_right = 0;
    var->blue.offset = 0;
    var->blue.length = 5;
    var->blue.msb_right = 0;
  }
  var->transp.offset = 0;
  var->transp.length = 0;
  var->transp.msb_right = 0;

  var->nonstd = 0;
  var->activate = 0;

  var->height = -1;
  var->width = -1;
  var->accel = mach64_use_accel __MN__  ;
  var->vmode = FB_VMODE_NONINTERLACED;

  /* Dummy values */

  if (par->bpp == 8)
    var->pixclock = M64_8_PIXCLOCK;
  else
    var->pixclock = M64_16_PIXCLOCK;
  var->sync = 0;
  var->left_margin = 64;
  var->right_margin = 96;
  var->upper_margin = 35;
  var->lower_margin = 12;
  var->hsync_len = 112;
  var->vsync_len = 2;

  for(i = 0; i < arraysize(var->reserved); i++)
    var->reserved[i] = 0;

  return(0);
}
/*endproc _encode_var() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Set a single color register. The values supplied are already
    *    rounded down to the hardware's capabilities (according to the
    *    entries in the var structure). Return != 0 for invalid regno.
    */
/*--------------------------------------------------------------------------*/

static int 
M64_hw_setcolreg(u_int regno,
  u_int red,
  u_int green,
  u_int blue,
  u_int transp)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 

  unsigned short                colmask;
  unsigned short                colshift;

  /*--------------------------------------------------------------------------*/
  OFuncId("M64_hw_setcolreg(%02X) \n", regno);

  if(regno > 255)  return (1);
  mach64_colormap  __MN__  [regno][0] = red;
  mach64_colormap  __MN__  [regno][1] = green;
  mach64_colormap  __MN__  [regno][2] = blue;
  mach64_colormap  __MN__  [regno][3] = transp;

  if(! phys_vgaiobase)  return 1;

  colmask = (1 << mach64_bitspercol __MN__  ) - 1;
  colshift = 8 - mach64_bitspercol __MN__  ;

  OUTB(0x3c8, regno);
  DACDelay;
  OUTB(0x3c9, (red >> colshift) & colmask);
  DACDelay;
  OUTB(0x3c9, (green >> colshift) & colmask);
  DACDelay;
  OUTB(0x3c9, (blue >> colshift) & colmask);
  DACDelay;

  return (0);
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Read a single color register and split it into
    *    colors/transparent. Return != 0 for invalid regno.
    */
/*--------------------------------------------------------------------------*/

static int 
M64_hw_getcolreg(u_int regno,
  u_int * red,
  u_int * green,
  u_int * blue,
  u_int * transp)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 

  /*--------------------------------------------------------------------------*/
  OFuncId("M64_hw_getcolreg() \n");

  if(regno >= 256)  return (1);

  *red    = mach64_colormap  __MN__  [regno][0];
  *green  = mach64_colormap  __MN__  [regno][1];
  *blue   = mach64_colormap  __MN__  [regno][2];
  *transp = mach64_colormap  __MN__  [regno][3];

  /*
  OUTB(0x3c7, regno);
  DACDelay;
  *red = INB(0x3c9) & colmask;
  DACDelay;
  *green = INB(0x3c9) & colmask;
  DACDelay;
  *blue = INB(0x3c9) & colmask;
  DACDelay;
  */
	    
  return (0);
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    (Un)Blank the screen: Set all colors = RGB 000 or back
    */
/*--------------------------------------------------------------------------*/

void 
M64_hw_blank(int blank)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 
  int regno;

  unsigned short                colmask;
  unsigned short                colshift;

  /*--------------------------------------------------------------------------*/
  CFuncId("M64_hw_blank() \n");

  if(blank)
  {
    for(regno = 0; regno < 256; regno++)
    {
      OUTB(0x3c8, regno);
      DACDelay;
      OUTB(0x3c9, 0);
      DACDelay;
      OUTB(0x3c9, 0);
      DACDelay;
      OUTB(0x3c9, 0);
      DACDelay;

    }
    /*nexxt regno*/
  }
  else
  {
    colmask = (1 << mach64_bitspercol __MN__  ) - 1;
    colshift = 8 - mach64_bitspercol __MN__  ;
    for (regno = 0; regno < 256; regno++)
    {
      OUTB(0x3c8, regno);
      DACDelay;
      OUTB(0x3c9, (mach64_colormap  __MN__  [regno][0] >> colshift) & colmask);
      DACDelay;
      OUTB(0x3c9, (mach64_colormap  __MN__  [regno][1] >> colshift) & colmask);
      DACDelay;
      OUTB(0x3c9, (mach64_colormap  __MN__  [regno][2] >> colshift) & colmask);
    }
    /*nexxt regno*/
  }
  /*endif*/

  return;
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
/*
 * We are waiting for "fifo" FIFO-slots empty
 */
/*--------------------------------------------------------------------------*/

void 
Mach64_WaitQueue (u_short fifo)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  CFuncId("Mach64_WaitQueue() \n");

  WaitQueue(16);

  return;
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
/*
 * We are waiting for Hardware (Graphics Engine) not busy
 */
/*--------------------------------------------------------------------------*/

void 
Mach64_WaitBlit (void)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  CFuncId("Mach64_WaitBlit() \n");

  WaitIdleEmpty(); /* Make sure that all commands have finished */

  return;
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
/*
 * BitBLT - Through the Plane
 */
/*--------------------------------------------------------------------------*/

void 
Mach64_BitBLT (
  int     xSrc,
  int     ySrc,
  int     xDst,
  int     yDst,
  u_int   wSrc,
  u_int   hSrc,
  u_short mode)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 
  u_int                         direction;

  /*--------------------------------------------------------------------------*/
  CFuncId("BB(%03X,%03X,%03X,%03X,%03X,%03X,%X) ",
    xSrc, ySrc, xDst, yDst, wSrc, hSrc, mode);
  /* Set drawing direction */
  /* -Y, X maj, -X (default) */

#if 1

  direction = 0;

  if(ySrc < yDst)
  {
    yDst += hSrc - 1;
    ySrc += hSrc - 1;
  }
  else
  {
    direction |= DST_Y_TOP_TO_BOTTOM;
  }
  /*endif*/
  if(xSrc < xDst)
  {
    xDst += wSrc - 1;
    xSrc += wSrc - 1;
  }
  else
  {
    direction |= DST_X_LEFT_TO_RIGHT;
  }
  /*endif*/


  WaitQueue(4);
  regw(DP_WRITE_MASK, 0x000000FF          /* pGC->planemask */ );
  regw(DP_MIX, (mach64alu[mode] << 16) |  MIX_DST);
  regw(DP_SRC, FRGD_SRC_BLIT);

  regw(GUI_TRAJ_CNTL, direction);

/*
  MACH64_BIT_BLT(xSrc, ySrc, xDst, yDst, wSrc, hSrc, direction);
*/
    WaitQueue(4);
    regw(SRC_Y_X, (((xSrc) << 16) | (ySrc) ));
    regw(SRC_WIDTH1, (wSrc));
    regw(DST_Y_X, (((xDst) << 16) | (yDst) ));
    regw(DST_HEIGHT_WIDTH, (((wSrc) << 16) | (hSrc) ));

  WaitIdleEmpty(); /* Make sure that all commands have finished */

  /*
   * Make sure that the destination trajectory is correctly set
   * for subsequent calls.  MACH64_BIT_BLT is the only function that
   * currently changes the destination trajectory from L->R and T->B.
   */
  regw(DST_CNTL, (DST_X_LEFT_TO_RIGHT | DST_Y_TOP_TO_BOTTOM));

#endif

  return;
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
/*
 * Rectangle Fill Solid
 */
/*--------------------------------------------------------------------------*/

void 
Mach64_RectFill(
  int     x,
  int     y,
  u_int   wSrc,
  u_int   hSrc,
  u_int   mode,
  u_int   color)

/*--------------------------------------------------------------------------*/
{

#if 1

  /*--------------------------------------------------------------------------*/
  CFuncId("RF(%03X,%03X,%03X,%03X,%X) ",
    x, y, wSrc, hSrc, mode);

  WaitQueue(5);
  regw(DP_FRGD_CLR, color              /* pGC->fgPixel */ );
  regw(DP_WRITE_MASK, 0x000000FF       /* pGC->planemask */ );
  regw(DP_MIX, (mach64alu[mode] << 16) | MIX_DST);
  regw(DP_SRC, FRGD_SRC_FRGD_CLR);

  regw(DST_CNTL, DST_X_LEFT_TO_RIGHT | DST_Y_TOP_TO_BOTTOM);

  if ((wSrc > 0) && (hSrc > 0))
  {
    /* Bprint("q\n"); */
    WaitQueue(2);
    regw(DST_Y_X, (((u_int)x << 16) | ((u_int)y & 0x0000ffff)));
    regw(DST_HEIGHT_WIDTH, (((u_int)wSrc << 16) | hSrc));
  }

  WaitIdleEmpty(); /* Make sure that all commands have finished */

#endif

  return;
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
/*
 * Move cursor to x, y
 */
/*--------------------------------------------------------------------------*/

void 
Mach64_MoveCursor (u_short x, u_short y)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  CFuncId("Mach64_MoveCursor() \n");


  return;
}
/*endproc Mach64_MoveCursor()*/

/*============================================================================*/


/* -------------------- Generic routines ------------------------------------ */


/*--------------------------------------------------------------------------*/
   /*
    *    Fill the hardware's `par' structure.
    */
/*--------------------------------------------------------------------------*/

static void 
M64_hw_get_par(struct mach64_fb_par * par)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 

  /*--------------------------------------------------------------------------*/
  CFuncId("M64_hw_get_par() \n");
  if(current_par_valid __MN__  )
  {
    *par = current_par __MN__  ;
  }
  else    /* use a default, initialize */
  {
    fbhws_ptr->decode_var(
      &mach64_fb_predefined[mach64_mode __MN__  ],
      par);
  }
  /*endif*/

  return;
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static void 
M64_hw_set_par(struct mach64_fb_par * par)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 

  /*--------------------------------------------------------------------------*/
  CFuncId("M64_hw_set_par() \n");
  current_par __MN__   = *par;
  current_par_valid __MN__   = 1;


  return;
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static int 
do_fb_set_var(struct fb_var_screeninfo * var, int isactive)

/*--------------------------------------------------------------------------*/
{
  int                           err, activate;
  struct mach64_fb_par          par;

  /*--------------------------------------------------------------------------*/
  CFuncId("do_fb_set_var() \n");
  if((err = fbhws_ptr->decode_var(var, &par)))
    return(err);
  activate = var->activate;
  if((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW && isactive)
  {
    M64_hw_set_par(&par);
  }
  /*endif*/
  fbhws_ptr->encode_var(var, &par);
  var->activate = activate;
  return(0);
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static void 
do_install_cmap(int con)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  CFuncId("do_install_cmap() \n");

  if(con != currcon)  return;

  if(fb_display[con].cmap.len)
  {
    fb_set_cmap(&fb_display[con].cmap,
      &fb_display[con].var,
      1,
      fbhws_ptr->setcolreg);
  }
  else
  {
    fb_set_cmap(fb_default_cmap(fb_display[con].var.bits_per_pixel),
      &fb_display[con].var,
      1,
      fbhws_ptr->setcolreg);
  }
  /*endif*/

  return;
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Get the Fixed Part of the Display
    */
/*--------------------------------------------------------------------------*/

static int 
M64_op_get_fix(struct fb_fix_screeninfo * fix, int con)

/*--------------------------------------------------------------------------*/
{
  struct mach64_fb_par          par;
  int                           status = 0;

  /*--------------------------------------------------------------------------*/
  CFuncId("M64_op_get_fix() \n");
  if (con == -1)
  {
    M64_hw_get_par(&par);
  }
  else
  {
    status = fbhws_ptr->decode_var(&fb_display[con].var, &par);
  }
  /*endif*/
  return(status ? status : fbhws_ptr->encode_fix(fix, &par));
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Get the User Defined Part of the Display
    */
/*--------------------------------------------------------------------------*/

static int 
M64_op_get_var(struct fb_var_screeninfo *var, int con)

/*--------------------------------------------------------------------------*/
{
  struct mach64_fb_par          par;
  int                           status = 0;

  /*--------------------------------------------------------------------------*/
  CFuncId("M64_op_get_var() \n");
  if(con == -1)
  {
    M64_hw_get_par(&par);
    status = fbhws_ptr->encode_var(var, &par);
  }
  else
  {
    *var = fb_display[con].var;
  }
  /*endif*/
  return(status);
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static void 
Mach64_fb_set_disp(int con)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 

  struct display *              disp_p;
  struct fb_fix_screeninfo      fix;

  /*--------------------------------------------------------------------------*/
  CFuncId("Mach64_fb_set_disp() \n");

  if(con >= 0)
    disp_p = &fb_display[con];
  else
    disp_p = &m_disp __MN__  ;  	/* used during initialization */

  M64_op_get_fix(&fix, con);

  if(con == -1)  con = 0;

  disp_p->screen_base = (u_char *)fix.smem_start;
  disp_p->visual = fix.visual;
  disp_p->type = fix.type;
  disp_p->type_aux = fix.type_aux;
  disp_p->ypanstep = fix.ypanstep;
  disp_p->ywrapstep = fix.ywrapstep;
  disp_p->line_length = fix.line_length;
  if(fix.visual != FB_VISUAL_PSEUDOCOLOR && fix.visual != FB_VISUAL_DIRECTCOLOR)
  {
    disp_p->can_soft_blank = 0;                   /* aus atari_ */
  }
  else
  {
    disp_p->can_soft_blank = 1;
  }
  /*endif*/

  disp_p->can_soft_blank = 1;                       /* aus Cyber_ */
  disp_p->inverse = mach64_inverse __MN__  ;

  return;
}
/*endproc Mach64_fb_set_disp() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/
  /*
   *    Set the User Defined Part of the Display
   */
/*--------------------------------------------------------------------------*/

static int 
M64_op_set_var(struct fb_var_screeninfo * var, int con)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 
  int       err, oldxres, oldyres, oldvxres, oldvyres, oldbpp;

  /*--------------------------------------------------------------------------*/
  CFuncId("M64_op_set_var() \n");

  if((err = do_fb_set_var(var, con == currcon)))
    return(err);
  if((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW)
  {
    oldxres = fb_display[con].var.xres;
    oldyres = fb_display[con].var.yres;
    oldvxres = fb_display[con].var.xres_virtual;
    oldvyres = fb_display[con].var.yres_virtual;
    oldbpp = fb_display[con].var.bits_per_pixel;
    fb_display[con].var = *var;
    if(oldxres != var->xres || oldyres != var->yres ||
      oldvxres != var->xres_virtual || oldvyres != var->yres_virtual ||
      oldbpp != var->bits_per_pixel)
    {
      Mach64_fb_set_disp(con);
      (*fb_info __MN__  .changevar)(con);
      fb_alloc_cmap(&fb_display[con].cmap, 0, 0);
      do_install_cmap(con);
    }
  }
  /*endif*/
  var->activate = 0;

  return(0);
}
/*endproc M64_op_set_var() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Get the Colormap
    */
/*--------------------------------------------------------------------------*/

static int 
M64_op_get_cmap(struct fb_cmap * cmap, int kspc, int con)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  OFuncId("M64_op_get_cmap() \n");

  if(con == currcon) /* current console? */
  {
    return(fb_get_cmap(cmap, &fb_display[con].var, kspc, fbhws_ptr->getcolreg));
  }
  else if(fb_display[con].cmap.len) /* non default colormap? */
  {
    fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
  }
  else
  {
    fb_copy_cmap(fb_default_cmap(fb_display[con].var.bits_per_pixel),
      cmap, kspc ? 0 : 2);
  }
  /*endif*/

  return(0);
}
/*endproc M64_op_get_cmap() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Set the Colormap
    */
/*--------------------------------------------------------------------------*/

static int 
M64_op_set_cmap(struct fb_cmap * cmap, int kspc, int con)

/*--------------------------------------------------------------------------*/
{
  int err;

  /*--------------------------------------------------------------------------*/
  OFuncId("M64_op_set_cmap() \n");

  if (!fb_display[con].cmap.len)       /* no colormap allocated? */
  {
    err = fb_alloc_cmap(&fb_display[con].cmap,
      1 << fb_display[con].var.bits_per_pixel,
      0);
    if(err)
    {
      return(err);
    }
    /*endif*/
  }
  /*endif*/
  if(con == currcon)              /* current console? */
  {
    return(fb_set_cmap(cmap, &fb_display[con].var, kspc, fbhws_ptr->setcolreg));
  }
  else
  {
    fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
  }
  /*endif*/

  return(0);
}
/*endproc M64_op_set_cmap() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Pan or Wrap the Display
    *
    *    This call looks only at xoffset, yoffset and the FB_VMODE_YWRAP flag
    */
/*--------------------------------------------------------------------------*/

static int 
M64_op_pan_display(struct fb_var_screeninfo *var, int con)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  CFuncId("Mach64_pan_display() \n");

  return(-EINVAL);
}
/*endproc M64_op_pan_display() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Mach 64 Frame Buffer Specific ioctls
    */
/*--------------------------------------------------------------------------*/

static int 
M64_op_ioctl(
  struct inode * inode,
  struct file *  file,
  u_int          cmd,
  u_long         arg,
  int            con)

/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 

  /*--------------------------------------------------------------------------*/
  CFuncId("M64_op_ioctl($%08X %d) \n", cmd, con);
  switch (cmd)
  {

#ifdef FBCMD_GET_CURRENTPAR
    case FBCMD_GET_CURRENTPAR:
    {
      if (copy_to_user((void *)arg, (void *)&current_par __MN__  ,
         sizeof(struct mach64_fb_par)))
        return -EFAULT;
      return 0;
    }
    /*endcase*/
#endif

#ifdef FBCMD_SET_CURRENTPAR
    case FBCMD_SET_CURRENTPAR:
    {

      if (copy_from_user((void *)&current_par __MN__  , (void *)arg,
        sizeof(struct mach64_fb_par)))
        return -EFAULT;
      M64_hw_set_par(&current_par __MN__  );
      return 0;
    }
    /*endcase*/
#endif

  }
  /*endswitch*/
  return(-EINVAL);
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/

int 
Mach64_probe(void)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  FuncId("Mach64_probe() \n");
  mach64_key = 0;
  return(mach64_key);
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/

__initfunc(
void 
Mach64_video_setup(char * options, int * ints)
)
/*--------------------------------------------------------------------------*/
{
  DECLARE_STATIC_UINT08___MN_ 

  char *        this_opt;
  char          mach64_str[80];

  /*--------------------------------------------------------------------------*/
  FuncId("Mach64_video_setup() '%s'\n", options);

  fb_info __MN__  .fontname[0] = '\0';

  if(!options || !*options)  return;

  for(this_opt = strtok(options, ","); this_opt; this_opt = strtok(NULL, ","))
  {
    if (!strcmp(this_opt, "inverse"))
    {
      mach64_inverse __MN__   = 1;
      fb_invert_cmaps();
    }
    else if (!strncmp(this_opt, "font:", 5))
    {
      strcpy(fb_info __MN__  .fontname, this_opt + 5);
    }
    else if (!strncmp(this_opt,"Mach64:", 7))
    {
      strncpy(mach64_str, this_opt + 7, 80);
      store_video_par(mach64_str  _ADD_ARG_MN_ );
      _INC_MN_ ;
      mach64_count =  _MN_ ;
    }
    else
    {
      mach64_mode __MN__   = get_video_mode(this_opt);
    }
    /*endif*/
  }
  /*next this_opt*/

  return;
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static int 
store_video_par(char* video_str ADD_ARG_UINT08___MN_ )

/*--------------------------------------------------------------------------*/
{
  char *                        p;
  unsigned  long                sm_addr, sm_size;

  int                           xres, yres, depth;
  short int                     pmode;
  short int                     use_accel;

  /*--------------------------------------------------------------------------*/
  CFuncId("store_video_par(%d) \n",  _MN_ );
  /* Format is: <xres>;<yres>;<depth>;<plane organ.>;
   *            <screen mem addr>
   *        [;<screen mem length>[;<vgaiobase>[;<colorreg-type>]]]
   */

  use_accel = 0;
  if (!(p = strtoke(video_str, ";")) || !*p)  goto mach64_invalid;
  xres = simple_strtoul(p, NULL, 10);
  if (xres <= 0)  goto mach64_invalid;

  if (!(p = strtoke(NULL, ";")) || !*p)  goto mach64_invalid;
  yres = simple_strtoul(p, NULL, 10);
  if (yres <= 0) goto mach64_invalid;

  if (!(p = strtoke(NULL, ";")) || !*p)  goto mach64_invalid;
  depth = simple_strtoul(p, NULL, 10);
  if (depth != 1 && depth != 2 && depth != 4 && depth != 8 &&
      depth != 16 && depth != 24)  goto mach64_invalid;

  if (!(p = strtoke(NULL, ";")) || !*p)  goto mach64_invalid;
  if (*p == 'i')
    pmode = FB_TYPE_INTERLEAVED_PLANES;
  else if (*p == 'p')
    pmode = FB_TYPE_PACKED_PIXELS;
  else if (*p == 'n')
    pmode = FB_TYPE_PLANES;
  else if (*p == 't')
    pmode = -1; /* true color */
  else
    goto mach64_invalid;


  if (!(p = strtoke(NULL, ";")) ||!*p)  goto mach64_invalid;
  sm_addr = simple_strtoul(p, NULL, 0);

  if (!(p = strtoke(NULL, ";")) ||!*p)
    sm_size = xres * yres * depth / 8;
  else
    sm_size = simple_strtoul(p, NULL, 0);

  if ((p = strtoke(NULL, ";")) && *p)
    phys_vgaiobase __MN__   = simple_strtoul(p, NULL, 0);

  if ((p = strtoke(NULL, ";")) && *p)
  {
    mach64_bitspercol __MN__   = simple_strtoul(p, NULL, 0);
    if (mach64_bitspercol __MN__   > 8)
      mach64_bitspercol __MN__   = 8;
    else if (mach64_bitspercol __MN__   < 1)
      mach64_bitspercol __MN__   = 1;
  }
  /*endif*/
  if ((p = strtoke(NULL, ";")) && *p)
  {
    if(*p == 'a')  use_accel = FB_ACCEL_MACH64;
  
  }
  /*endif*/

  mach64_xres __MN__    = xres;
  mach64_yres __MN__    = yres;
  mach64_depth __MN__   = depth;
  mach64_pmode __MN__   = pmode;
  mach64_use_accel __MN__   = use_accel;
  phys_vmembase __MN__   = sm_addr;
  phys_size __MN__    = sm_size;

  return(0);

  /*--------------------------------------------------------------------------*/
  mach64_invalid:

  phys_vmembase __MN__   = 0;

  return(-1);
}
/*endproc*/

/*============================================================================*/


/* a strtok which returns empty strings, too */

/*--------------------------------------------------------------------------*/

static char * 
strtoke(char * s, const char * ct)

/*--------------------------------------------------------------------------*/
{
  static char    *ssave = NULL;
  char          *sbegin, *send;
  
  /*--------------------------------------------------------------------------*/
  sbegin  = s ? s : ssave;
  if (!sbegin) {
    return NULL;
  }
  if (*sbegin == '\0') {
    ssave = NULL;
    return NULL;
  }
  send = strpbrk(sbegin, ct);
  if (send && *send != '\0')
    *send++ = '\0';
  ssave = send;
  return sbegin;
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Initialization
    */
/*--------------------------------------------------------------------------*/

__initfunc(
int
Mach64_fb_init(u_long * mem_start)
)
/*--------------------------------------------------------------------------*/
{
  static u_char                 fb_init_count;
  DECLARE_STATIC_UINT08___MN_

  struct mach64_fb_par          par;
  int                           err;
  DECLARE_UINT08___MN_

  /*--------------------------------------------------------------------------*/
  FuncId("Mach64_fb_init($%08lX) \n", *mem_start);

  /*
  _MN_ = fb_init_count;
  */
  if(phys_vmembase __MN__   == 0)  return -ENODEV;

  fb_init_count++;
  memstart = mem_start;

  fbhws_ptr = &mach64_switch;

  /* Map the video memory (physical address given) to somewhere
   * in the kernel address space.
   */
  *mem_start = (*mem_start + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1);
  virt_vmembase __MN__   = kernel_map(
    phys_vmembase __MN__  ,
    phys_size __MN__  ,
#ifdef  _OHNE_HW_ACCEL_
    KERNELMAP_NO_COPYBACK,    /* must be changed if hardware blit */
#else
    KERNELMAP_NOCACHE_SER,
#endif
    mem_start);

  if (phys_vgaiobase __MN__  )
  {

#ifdef _MAP_768K_IO_

    virt_vgaiobase __MN__   = kernel_map(
      phys_vgaiobase __MN__  - 0x00080000,
      0x0C0000,                          /* 768 K */
      KERNELMAP_NOCACHE_SER,
      mem_start);

#else

    virt_vgaiobase __MN__   = kernel_map(
      phys_vgaiobase __MN__  ,
      0x10000,                           /* 64 K */
      KERNELMAP_NOCACHE_SER,
      mem_start);

#ifdef  _MAP_GUIREGS_0BFC00_
    phys_mmregbase __MN__   = phys_vgaiobase __MN__  - 0x00100000 + 0x000B0000;
    Dprint(" phys_mmregbase=$%08lX \n", phys_mmregbase __MN__  );
    
    virt_mmregbase __MN__   = kernel_map(
      phys_mmregbase __MN__  ,
      0x10000,                           /* 64 K */
      KERNELMAP_NOCACHE_SER,
      mem_start);
#endif
    
#endif  /* _MAP_768K_IO_ */
  }
  /*endif*/

  virt_size __MN__   = phys_size __MN__   & PAGE_MASK;  /* abrunden */
  virt_size __MN__   = (phys_size __MN__  - 0x00200000) & PAGE_MASK;  /* abrunden */

  Dprint(" PAGE_MASK=$%08lX phys_size=$%08lX \n", PAGE_MASK, phys_size __MN__ );
  /*
  memset((char *) virt_vmembase __MN__  , 0, virt_size __MN__  );
  */

  strcpy(fb_info __MN__  .modename, mach64_fb_name);
  fb_info __MN__  .changevar = NULL;
  fb_info __MN__  .node = -1;
  fb_info __MN__  .fbops = &mach64_fb_ops;
  fb_info __MN__  .fbvar_num = NUM_TOTAL_MODES;
  fb_info __MN__  .fbvar = mach64_fb_predefined;
  fb_info __MN__  .disp = &m_disp __MN__  ;
  fb_info __MN__  .switch_con = &Mach64fb_switch;
  fb_info __MN__  .updatevar = &Mach64fb_updatevar;
  fb_info __MN__  .blank = &Mach64fb_blank;
  fb_info __MN__  .setcmap = &Mach64fb_setcmap;
  strcat(fb_info __MN__  .modename, fb_var_names[mach64_mode __MN__  ]);
  fb_info __MN__  .modename[39] =  _MN_ ;

  err = register_framebuffer(&fb_info __MN__  );
  if(err < 0)
  {
    Eprint(" Cannot register frame buffer() \n" );
    return(err);
  }
  /*endif*/

  fbhws_ptr->init( _ARG_MN_ );
  fbhws_ptr->decode_var(&mach64_fb_predefined[mach64_mode __MN__  ], &par);
  fbhws_ptr->encode_var(&mach64_fb_predefined[0], &par); /* store resolution */

  do_fb_set_var(&mach64_fb_predefined[0], 1);

  M64_op_get_var(&m_disp __MN__  .var, -1);  /* fill in m_disp.var */
  Mach64_fb_set_disp(-1);                    /* fill in the rest of m_disp  */

  Iprint(" Determined %dx%d, depth %d $%08lX MEM @ $%08lX IO @ $%08lX\n",
    m_disp __MN__  .var.xres,
    m_disp __MN__  .var.yres,
    m_disp __MN__  .var.bits_per_pixel,
    phys_size __MN__  ,
    phys_vmembase __MN__  ,
    phys_vgaiobase __MN__  );
  do_install_cmap(0);

  Iprint(
    " %s FB device %d, using %ldK of video MEM @ $%08lX IO @ $%08lX $%08lX\n",
    fb_info __MN__  .modename,
     _MN_ ,
    virt_size __MN__   >> 10,
    virt_vmembase __MN__  ,
    virt_vgaiobase __MN__  ,
    virt_mmregbase __MN__  );

  _INC_MN_ ;

  return(0);
}
/*endproc Mach64_fb_init() */

/*============================================================================*/


/*--------------------------------------------------------------------------*/

static int 
Mach64fb_switch(int con)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  CFuncId("Mach64fb_switch(%d) \n", con);
  /* Do we have to save the colormap? */
  if (fb_display[currcon].cmap.len)
  {
    fb_get_cmap(&fb_display[currcon].cmap,
      &fb_display[currcon].var,
      1,
      fbhws_ptr->getcolreg);
  }
  /*endif*/
  do_fb_set_var(&fb_display[con].var, 1);
  currcon = con;
  /* Install new colormap */
  do_install_cmap(con);
  return(0);
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Update the `var' structure (called by fbcon.c)
    *
    *    This call looks only at yoffset and the FB_VMODE_YWRAP flag in `var'.
    *    Since it's called by a kernel driver, no range checking is done.
    */
/*--------------------------------------------------------------------------*/

static int 
Mach64fb_updatevar(int con)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  CFuncId("Mach64fb_updatevar() \n");
  return(0);
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Blank the display.
    */
/*--------------------------------------------------------------------------*/

static void 
Mach64fb_blank(int blank)

/*--------------------------------------------------------------------------*/
{
  /*--------------------------------------------------------------------------*/
  CFuncId("Mach64fb_blank() \n");
  fbhws_ptr->blank(blank);
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Set the colormap
    */
/*--------------------------------------------------------------------------*/

static int 
Mach64fb_setcmap(struct fb_cmap *cmap, int con)

/*--------------------------------------------------------------------------*/
{

  /*--------------------------------------------------------------------------*/
  CFuncId("Mach64fb_setcmap() \n");
  return(M64_op_set_cmap(cmap, 1, con));
}
/*endproc*/

/*============================================================================*/


/*--------------------------------------------------------------------------*/
   /*
    *    Get a Video Mode
    */
/*--------------------------------------------------------------------------*/

static int 
get_video_mode(const char * name)

/*--------------------------------------------------------------------------*/
{
  int i;

  /*--------------------------------------------------------------------------*/
  CFuncId("get_video_mode(%s) \n", name);

  for (i = 1; i < NUM_PREDEF_MODES; i++)
  {
    if (!strcmp(name, fb_var_names[i]))  return(i);
  }
  /*next i*/
  return(0);
}
/*endproc*/

/*============================================================================*/


/*============================================================================*/
/* RCS-ID */

  static char
  rcs_id_c[] =
    "@(#) Linux-m68k: 1997 08 16 20:54:00 BHA: mach64fb.c ";

  static char
  rcs_id_h[] = MACH64FB_H_ID;

  static char
  rcs_id_regs_h[] = MACH64REGS_H_ID;

/*============================================================================*/
